"""Tests for workflow_explainer."""

from yagra.application.use_cases.workflow_explainer import explain_workflow
from yagra.domain.entities.graph_schema import GraphSpec


def _make_spec(overrides: dict) -> GraphSpec:
    base = {
        "version": "1",
        "start_at": "node_a",
        "end_at": ["node_a"],
        "nodes": [
            {
                "id": "node_a",
                "handler": "llm",
                "params": {"model": "gpt-4o-mini", "prompt": "hello"},
            }
        ],
        "edges": [],
    }
    base.update(overrides)
    return GraphSpec.model_validate(base)


def test_explain_entry_and_exit():
    spec = _make_spec({})
    result = explain_workflow(spec)
    assert result["entry_point"] == "node_a"
    assert result["exit_points"] == ["node_a"]


def test_explain_required_handlers():
    spec = _make_spec({})
    result = explain_workflow(spec)
    assert "llm" in result["required_handlers"]


def test_explain_variable_flow_extracts_prompt_vars():
    spec = _make_spec(
        {
            "nodes": [
                {
                    "id": "node_a",
                    "handler": "llm",
                    "params": {
                        "model": "gpt-4o-mini",
                        "prompt": "translate {text} to {lang}",
                        "output_key": "translation",
                    },
                }
            ],
        }
    )
    result = explain_workflow(spec)
    assert result["variable_flow"]["node_a"]["inputs"] == ["text", "lang"]
    assert "translation" in result["variable_flow"]["node_a"]["outputs"]


def test_explain_conditional_node_has_next_in_outputs():
    spec = GraphSpec.model_validate(
        {
            "version": "1",
            "start_at": "classify",
            "end_at": ["approve", "reject"],
            "nodes": [
                {
                    "id": "classify",
                    "handler": "llm",
                    "params": {"model": "gpt-4o-mini", "prompt": "classify {input}"},
                },
                {
                    "id": "approve",
                    "handler": "llm",
                    "params": {"model": "gpt-4o-mini", "prompt": "approved"},
                },
                {
                    "id": "reject",
                    "handler": "llm",
                    "params": {"model": "gpt-4o-mini", "prompt": "rejected"},
                },
            ],
            "edges": [
                {"source": "classify", "target": "approve", "condition": "approved"},
                {"source": "classify", "target": "reject", "condition": "rejected"},
            ],
        }
    )
    result = explain_workflow(spec)
    assert "__next__" in result["variable_flow"]["classify"]["outputs"]


def test_explain_execution_paths_linear():
    spec = GraphSpec.model_validate(
        {
            "version": "1",
            "start_at": "a",
            "end_at": ["b"],
            "nodes": [
                {"id": "a", "handler": "llm", "params": {"model": "gpt-4o-mini", "prompt": "x"}},
                {"id": "b", "handler": "llm", "params": {"model": "gpt-4o-mini", "prompt": "y"}},
            ],
            "edges": [{"source": "a", "target": "b"}],
        }
    )
    result = explain_workflow(spec)
    assert result["execution_paths"] == [["a", "b"]]


def test_explain_execution_paths_branch():
    spec = GraphSpec.model_validate(
        {
            "version": "1",
            "start_at": "classify",
            "end_at": ["approve", "reject"],
            "nodes": [
                {
                    "id": "classify",
                    "handler": "llm",
                    "params": {"model": "gpt-4o-mini", "prompt": "c"},
                },
                {
                    "id": "approve",
                    "handler": "llm",
                    "params": {"model": "gpt-4o-mini", "prompt": "a"},
                },
                {
                    "id": "reject",
                    "handler": "llm",
                    "params": {"model": "gpt-4o-mini", "prompt": "r"},
                },
            ],
            "edges": [
                {"source": "classify", "target": "approve", "condition": "approved"},
                {"source": "classify", "target": "reject", "condition": "rejected"},
            ],
        }
    )
    result = explain_workflow(spec)
    paths = result["execution_paths"]
    assert ["classify", "approve"] in paths
    assert ["classify", "reject"] in paths



"""プロンプトテンプレート変数抽出ユーティリティを提供する。"""

from __future__ import annotations

import re
from typing import Any

_LLM_HANDLERS = frozenset({"llm", "streaming_llm", "structured_llm"})


def _extract_required_vars(params: dict[str, Any]) -> list[str]:
    """ノード params からプロンプトが要求する変数名リストを返す。

    ``input_keys`` が明示指定されていればそれを優先し、
    未指定の場合はユーザープロンプトテンプレートから ``{variable}`` を自動抽出する。
    ハンドラーと同一のロジックを使用する。

    Args:
        params: ノードの params 辞書。

    Returns:
        要求する変数名のリスト。
    """
    explicit_keys = params.get("input_keys")
    if explicit_keys is not None:
        if isinstance(explicit_keys, list):
            return [str(k) for k in explicit_keys]
        return []

    prompt = params.get("prompt")
    if not isinstance(prompt, dict):
        return []
    user_template = prompt.get("user", "")
    if not isinstance(user_template, str):
        return []
    return re.findall(r"\{(\w+)\}", user_template)


def _get_output_key(params: dict[str, Any]) -> str:
    """ノード params から output_key を返す。未指定時はデフォルト ``"output"``。

    Args:
        params: ノードの params 辞書。

    Returns:
        output_key 文字列。
    """
    key = params.get("output_key", "output")
    return str(key) if key else "output"



r"""YAML スキーマ定義から Pydantic BaseModel を動的に生成するユーティリティ。

WebUI の Schema Settings テキストエリアに入力された YAML を
Pydantic モデルに変換し、structured_llm ハンドラーで使用する。

YAML フォーマット例::

    name: str
    age: int
    score: float
    tags: list[str]

Examples:
    基本的な使用方法:

    >>> from yagra.handlers.schema_builder import build_model_from_schema_yaml
    >>> schema_yaml = "name: str\\nage: int"
    >>> Model = build_model_from_schema_yaml(schema_yaml)
    >>> instance = Model(name="Alice", age=30)
    >>> print(instance)
    name='Alice' age=30
"""

from typing import Any

import yaml
from pydantic import BaseModel, create_model

# 安全な型マッピング辞書。eval() を使用せず、
# ホワイトリスト方式で型文字列を Python 型に解決する。
TYPE_MAP: dict[str, Any] = {
    # プリミティブ型
    "str": str,
    "int": int,
    "float": float,
    "bool": bool,
    # コレクション型
    "list[str]": list[str],
    "list[int]": list[int],
    "list[float]": list[float],
    "list[bool]": list[bool],
    # 辞書型
    "dict[str, str]": dict[str, str],
    "dict[str, int]": dict[str, int],
    "dict[str, Any]": dict[str, Any],
    # Optional 型
    "str | None": str | None,
    "int | None": int | None,
    "float | None": float | None,
    "bool | None": bool | None,
}


class SchemaYamlError(ValueError):
    """schema_yaml のパースまたはモデル生成に失敗した場合の例外。

    YAML 構文エラー、未サポートの型指定、不正なフォーマットなど、
    動的スキーマ生成に関連するすべてのエラーをこの例外で表現する。
    """


def build_model_from_schema_yaml(
    schema_yaml: str,
    model_name: str = "DynamicSchema",
) -> type[BaseModel]:
    r"""YAML 文字列から Pydantic BaseModel サブクラスを動的に生成する。

    フラットな ``key: type`` 形式の YAML を受け取り、各フィールドの型を
    安全な TYPE_MAP で解決して ``pydantic.create_model()`` でモデルを生成する。

    入力例::

        name: str
        age: int
        tags: list[str]
        nickname: str | None

    Args:
        schema_yaml: YAML 形式のスキーマ定義文字列（フラットな key: type 形式）。
        model_name: 生成するモデルのクラス名。デフォルトは ``"DynamicSchema"``。

    Returns:
        動的に生成された Pydantic BaseModel サブクラス。

    Raises:
        SchemaYamlError: YAML パースエラー、未サポート型、空フィールド、
            不正なフォーマットの場合に送出する。

    Examples:
        >>> Model = build_model_from_schema_yaml("name: str\\nage: int")
        >>> instance = Model.model_validate({"name": "Alice", "age": 30})
        >>> instance.name
        'Alice'
    """
    # 1. YAML パース
    try:
        parsed = yaml.safe_load(schema_yaml)
    except yaml.YAMLError as e:
        msg = f"Invalid YAML syntax in schema_yaml: {e}"
        raise SchemaYamlError(msg) from e

    # 2. パース結果のバリデーション
    if not isinstance(parsed, dict):
        msg = "schema_yaml must be a YAML mapping of field_name: type (e.g. 'name: str')"
        raise SchemaYamlError(msg)

    if len(parsed) == 0:
        msg = "schema_yaml must define at least one field"
        raise SchemaYamlError(msg)

    # 3. 各フィールドの型解決と Pydantic フィールド定義の構築
    field_definitions: dict[str, Any] = {}
    supported_types = ", ".join(sorted(TYPE_MAP.keys()))

    for field_name, type_str in parsed.items():
        if not isinstance(field_name, str):
            msg = f"Field name must be a string, got {type(field_name).__name__}: {field_name}"
            raise SchemaYamlError(msg)

        if not isinstance(type_str, str):
            msg = (
                f"Type for field '{field_name}' must be a string, "
                f"got {type(type_str).__name__}: {type_str}"
            )
            raise SchemaYamlError(msg)

        type_str_normalized = type_str.strip()

        resolved_type = TYPE_MAP.get(type_str_normalized)
        if resolved_type is None:
            msg = (
                f"Unsupported type '{type_str_normalized}' for field '{field_name}'. "
                f"Supported types: {supported_types}"
            )
            raise SchemaYamlError(msg)

        # Optional 型の場合はデフォルト値を None に設定
        if "| None" in type_str_normalized:
            field_definitions[field_name] = (resolved_type, None)
        else:
            field_definitions[field_name] = (resolved_type, ...)

    # 4. pydantic.create_model() でモデル生成
    return create_model(model_name, **field_definitions)



"""ワークフロー YAML を静的解析して実行情報を返す。"""

from __future__ import annotations

import re
from typing import Any

from yagra.domain.entities.graph_schema import GraphSpec, NodeSpec


def explain_workflow(spec: GraphSpec) -> dict[str, Any]:
    """GraphSpec を静的解析して実行情報を返す。

    ワークフローを実際に実行せずに解析し、エントリポイント・終了ノード・
    実行パス・必要ハンドラー・変数フローを返す。

    Args:
        spec: 解析対象の GraphSpec。

    Returns:
        以下のキーを持つ辞書:
        - entry_point: start_at ノード名
        - exit_points: end_at ノード名リスト
        - execution_paths: 可能な実行パス一覧（各パスはノード名リスト）
        - required_handlers: 必要なハンドラー名の重複なしリスト
        - variable_flow: ノード名をキーとする入出力変数マップ
    """
    node_map = {node.id: node for node in spec.nodes}

    return {
        "entry_point": spec.start_at,
        "exit_points": list(spec.end_at),
        "execution_paths": _enumerate_paths(spec),
        "required_handlers": _collect_handlers(spec),
        "variable_flow": _build_variable_flow(spec, node_map),
    }


def _enumerate_paths(spec: GraphSpec) -> list[list[str]]:
    """グラフを DFS で辿り、可能な実行パスを列挙する。

    条件分岐がある場合はすべての分岐パスを列挙する。
    ループが検出された場合は最初の再訪問で打ち切る（無限ループ防止）。

    Args:
        spec: 解析対象の GraphSpec。

    Returns:
        各実行パス（ノード名リスト）の一覧。
    """
    end_at_set = set(spec.end_at)

    # source → [target] のマッピングを構築
    adjacency: dict[str, list[str]] = {}
    for edge in spec.edges:
        adjacency.setdefault(edge.source, []).append(edge.target)

    paths: list[list[str]] = []

    def dfs(current: str, path: list[str], visited: set[str]) -> None:
        path = [*path, current]
        if current in end_at_set:
            paths.append(path)
            return
        next_nodes = adjacency.get(current, [])
        if not next_nodes:
            # エッジなしで終端でないノード（孤立ノード）はそのままパスとして記録
            paths.append(path)
            return
        for next_node in next_nodes:
            if next_node in visited:
                # ループ検出: 打ち切り
                paths.append([*path, f"...(loop:{next_node})"])
                continue
            dfs(next_node, path, visited | {current})

    dfs(spec.start_at, [], set())
    return paths


def _collect_handlers(spec: GraphSpec) -> list[str]:
    """ワークフロー内で使用されているハンドラー名を重複なしで返す。

    Args:
        spec: 解析対象の GraphSpec。

    Returns:
        使用ハンドラー名の重複なしリスト（登場順）。
    """
    seen: set[str] = set()
    handlers: list[str] = []
    for node in spec.nodes:
        if node.handler not in seen:
            seen.add(node.handler)
            handlers.append(node.handler)
    return handlers


def _build_variable_flow(
    spec: GraphSpec, node_map: dict[str, NodeSpec]
) -> dict[str, dict[str, list[str]]]:
    """各ノードの入力変数・出力変数を抽出して返す。

    入力変数はプロンプトテンプレートの {変数名} から抽出する。
    出力変数は output_key パラメータから取得し、省略時は 'output' とする。
    条件分岐ノード（conditional edges の source）は '__next__' を出力変数に追加する。

    Args:
        spec: 解析対象の GraphSpec。
        node_map: ノード ID から NodeSpec へのマッピング。

    Returns:
        ノード名をキーとする辞書。値は {"inputs": [...], "outputs": [...]} の辞書。
    """
    # conditional edges の source ノードを特定
    conditional_sources = {edge.source for edge in spec.edges if edge.condition is not None}

    flow: dict[str, dict[str, list[str]]] = {}
    for node in spec.nodes:
        inputs = _extract_input_variables(node)
        outputs = _extract_output_variables(node, conditional_sources)
        flow[node.id] = {"inputs": inputs, "outputs": outputs}
    return flow


def _extract_input_variables(node: NodeSpec) -> list[str]:
    """ノードのプロンプトから入力変数を抽出する。

    Args:
        node: 入力変数を抽出する NodeSpec。

    Returns:
        プロンプトテンプレート内の {変数名} リスト（重複なし、登場順）。
    """
    params = node.params
    prompt = params.get("prompt") or params.get("prompt_ref")
    if prompt is None:
        return []

    # prompt が文字列の場合はそのまま解析
    if isinstance(prompt, str):
        return _extract_vars_from_text(prompt)

    # prompt が dict の場合は content フィールドを解析
    if isinstance(prompt, dict):
        content = prompt.get("content", "")
        if isinstance(content, str):
            return _extract_vars_from_text(content)
        return []

    # prompt が list の場合は各メッセージの content を解析
    if isinstance(prompt, list):
        vars_seen: set[str] = set()
        vars_list: list[str] = []
        for msg in prompt:
            if isinstance(msg, dict):
                content = msg.get("content", "")
                if isinstance(content, str):
                    for var in _extract_vars_from_text(content):
                        if var not in vars_seen:
                            vars_seen.add(var)
                            vars_list.append(var)
        return vars_list

    # prompt_ref はファイルパスのため静的解析では変数抽出不可
    return []


def _extract_vars_from_text(text: str) -> list[str]:
    """テキスト内の {変数名} パターンを抽出する。

    Args:
        text: 変数を抽出するテキスト。

    Returns:
        変数名リスト（重複なし、登場順）。
    """
    seen: set[str] = set()
    result: list[str] = []
    for match in re.finditer(r"\{(\w+)\}", text):
        var = match.group(1)
        if var not in seen:
            seen.add(var)
            result.append(var)
    return result


def _extract_output_variables(node: NodeSpec, conditional_sources: set[str]) -> list[str]:
    """ノードの出力変数を抽出する。

    output_key が明示指定されていればそれを使い、なければ 'output' を使う。
    条件分岐ノードの場合は '__next__' を追加する。

    Args:
        node: 出力変数を抽出する NodeSpec。
        conditional_sources: conditional edge の source ノード ID セット。

    Returns:
        出力変数名リスト。
    """
    outputs: list[str] = []
    output_key = node.params.get("output_key")
    if output_key:
        outputs.append(output_key)
    # output_key 未指定でもハンドラーがある場合は 'output' がデフォルト
    # （ただし custom ハンドラーは不明なので組み込みのみ判定）
    builtin_handlers = {"llm", "structured_llm", "streaming_llm"}
    if not output_key and node.handler in builtin_handlers:
        outputs.append("output")

    if node.id in conditional_sources:
        outputs.append("__next__")

    return outputs



"""Tests for MCP server tools (without actual MCP connection)."""


def test_tool_validate_workflow_valid():
    from yagra.adapters.inbound.mcp_server import _tool_validate_workflow

    yaml_content = """
version: "1"
start_at: translate
end_at:
  - translate
nodes:
  - id: translate
    handler: llm_handler
edges: []
"""
    result = _tool_validate_workflow(yaml_content)
    assert result["is_valid"] is True
    assert result["issues"] == []


def test_tool_validate_workflow_invalid_node_ref():
    from yagra.adapters.inbound.mcp_server import _tool_validate_workflow

    yaml_content = """
version: "1"
start_at: translat
end_at:
  - translate
nodes:
  - id: translate
    handler: llm_handler
edges: []
"""
    result = _tool_validate_workflow(yaml_content)
    assert result["is_valid"] is False
    assert len(result["issues"]) > 0


def test_tool_validate_workflow_yaml_error():
    from yagra.adapters.inbound.mcp_server import _tool_validate_workflow

    result = _tool_validate_workflow("not: valid: yaml: [")
    assert result["is_valid"] is False


def test_tool_explain_workflow():
    from yagra.adapters.inbound.mcp_server import _tool_explain_workflow

    yaml_content = """
version: "1"
start_at: translate
end_at:
  - translate
nodes:
  - id: translate
    handler: llm
    params:
      output_key: translation
edges: []
"""
    result = _tool_explain_workflow(yaml_content)
    assert result["entry_point"] == "translate"
    assert "llm" in result["required_handlers"]
    assert "translation" in result["variable_flow"]["translate"]["outputs"]


def test_tool_list_templates():
    from yagra.adapters.inbound.mcp_server import _tool_list_templates

    result = _tool_list_templates()
    assert "templates" in result
    assert isinstance(result["templates"], list)


def test_tool_list_handlers():
    from yagra.adapters.inbound.mcp_server import _tool_list_handlers

    result = _tool_list_handlers()
    assert "handlers" in result
    names = [h["name"] for h in result["handlers"]]
    assert "llm" in names
    assert "structured_llm" in names
    assert "streaming_llm" in names


def test_create_mcp_server_requires_mcp_package():
    """Mcp パッケージが未インストールの場合 ImportError が発生することを確認する。"""
    # このテストは mcp がインストールされていれば server オブジェクトが返ることを確認
    # mcp がなければ ImportError が出る（それも正しい動作）
    try:
        from yagra.adapters.inbound.mcp_server import create_mcp_server

        server = create_mcp_server()
        assert server is not None
    except ImportError:
        # mcp 未インストールは正常ケース
        pass



"""Unit tests for schema_builder module.

YAML スキーマ定義から Pydantic モデルを動的生成する
schema_builder モジュールの単体テスト。
"""

import pytest
from pydantic import BaseModel, ValidationError

from yagra.handlers.schema_builder import (
    TYPE_MAP,
    SchemaYamlError,
    build_model_from_schema_yaml,
)


class TestBuildModelFromSchemaYaml:
    """build_model_from_schema_yaml のテスト."""

    # --- 正常系 ---

    def test_basic_str_field(self) -> None:
        """`str` フィールドのモデルが正しく生成されること."""
        model = build_model_from_schema_yaml("name: str")
        instance = model(name="Alice")
        assert instance.name == "Alice"  # type: ignore[attr-defined]

    def test_basic_int_field(self) -> None:
        """`int` フィールドのモデルが正しく生成されること."""
        model = build_model_from_schema_yaml("age: int")
        instance = model(age=30)
        assert instance.age == 30  # type: ignore[attr-defined]

    def test_basic_float_field(self) -> None:
        """`float` フィールドのモデルが正しく生成されること."""
        model = build_model_from_schema_yaml("score: float")
        instance = model(score=3.14)
        assert instance.score == pytest.approx(3.14)  # type: ignore[attr-defined]

    def test_basic_bool_field(self) -> None:
        """`bool` フィールドのモデルが正しく生成されること."""
        model = build_model_from_schema_yaml("active: bool")
        instance = model(active=True)
        assert instance.active is True  # type: ignore[attr-defined]

    def test_multiple_fields(self) -> None:
        """複数フィールドのモデルが正しく生成されること."""
        schema_yaml = "name: str\nage: int\nscore: float"
        model = build_model_from_schema_yaml(schema_yaml)
        instance = model(name="Bob", age=25, score=4.5)
        assert instance.name == "Bob"  # type: ignore[attr-defined]
        assert instance.age == 25  # type: ignore[attr-defined]
        assert instance.score == pytest.approx(4.5)  # type: ignore[attr-defined]

    def test_list_str_field(self) -> None:
        """list[str] フィールドが正しく動作すること."""
        model = build_model_from_schema_yaml("tags: list[str]")
        instance = model(tags=["python", "ai"])
        assert instance.tags == ["python", "ai"]  # type: ignore[attr-defined]

    def test_list_int_field(self) -> None:
        """list[int] フィールドが正しく動作すること."""
        model = build_model_from_schema_yaml("scores: list[int]")
        instance = model(scores=[1, 2, 3])
        assert instance.scores == [1, 2, 3]  # type: ignore[attr-defined]

    def test_dict_str_str_field(self) -> None:
        """dict[str, str] フィールドが正しく動作すること."""
        model = build_model_from_schema_yaml("metadata: dict[str, str]")
        instance = model(metadata={"key": "value"})
        assert instance.metadata == {"key": "value"}  # type: ignore[attr-defined]

    def test_dict_str_int_field(self) -> None:
        """dict[str, int] フィールドが正しく動作すること."""
        model = build_model_from_schema_yaml("counts: dict[str, int]")
        instance = model(counts={"a": 1, "b": 2})
        assert instance.counts == {"a": 1, "b": 2}  # type: ignore[attr-defined]

    def test_optional_str_field(self) -> None:
        """`str | None` フィールドがデフォルト None になること."""
        model = build_model_from_schema_yaml("nickname: str | None")
        # デフォルト値で生成可能
        instance = model()
        assert instance.nickname is None  # type: ignore[attr-defined]
        # 値を指定して生成
        instance2 = model(nickname="Bob")
        assert instance2.nickname == "Bob"  # type: ignore[attr-defined]

    def test_optional_int_field(self) -> None:
        """`int | None` フィールドがデフォルト None になること."""
        model = build_model_from_schema_yaml("count: int | None")
        instance = model()
        assert instance.count is None  # type: ignore[attr-defined]

    def test_all_supported_types(self) -> None:
        """全サポート型でモデルが生成できること."""
        for type_str in TYPE_MAP:
            schema_yaml = f"field: {type_str}"
            model = build_model_from_schema_yaml(schema_yaml)
            assert issubclass(model, BaseModel)

    def test_model_validate_works(self) -> None:
        """model_validate が正しく動作すること."""
        model = build_model_from_schema_yaml("name: str\nage: int")
        instance = model.model_validate({"name": "Alice", "age": 30})
        assert instance.name == "Alice"  # type: ignore[attr-defined]
        assert instance.age == 30  # type: ignore[attr-defined]

    def test_model_json_schema_returns_valid_schema(self) -> None:
        """model_json_schema() が妥当な JSON Schema を返すこと."""
        model = build_model_from_schema_yaml("name: str\nage: int")
        json_schema = model.model_json_schema()
        assert json_schema["type"] == "object"
        assert "name" in json_schema["properties"]
        assert "age" in json_schema["properties"]
        assert "name" in json_schema["required"]
        assert "age" in json_schema["required"]

    def test_custom_model_name(self) -> None:
        """model_name を指定するとクラス名に反映されること."""
        model = build_model_from_schema_yaml("name: str", model_name="PersonInfo")
        assert model.__name__ == "PersonInfo"

    def test_default_model_name(self) -> None:
        """デフォルトのモデル名が DynamicSchema であること."""
        model = build_model_from_schema_yaml("name: str")
        assert model.__name__ == "DynamicSchema"

    def test_whitespace_in_type_is_trimmed(self) -> None:
        """型文字列の前後の空白がトリムされること."""
        model = build_model_from_schema_yaml("name:  str ")
        instance = model(name="Alice")
        assert instance.name == "Alice"  # type: ignore[attr-defined]

    # --- 異常系 ---

    def test_invalid_yaml_syntax(self) -> None:
        """不正な YAML 構文で SchemaYamlError が送出されること."""
        with pytest.raises(SchemaYamlError, match="Invalid YAML syntax"):
            build_model_from_schema_yaml(":\ninvalid: [yaml")

    def test_non_dict_yaml(self) -> None:
        """YAML のパース結果が dict でない場合に SchemaYamlError が送出されること."""
        with pytest.raises(SchemaYamlError, match="must be a YAML mapping"):
            build_model_from_schema_yaml("- item1\n- item2")

    def test_scalar_yaml(self) -> None:
        """YAML がスカラー値の場合に SchemaYamlError が送出されること."""
        with pytest.raises(SchemaYamlError, match="must be a YAML mapping"):
            build_model_from_schema_yaml("just a string")

    def test_empty_dict(self) -> None:
        """空の dict で SchemaYamlError が送出されること."""
        with pytest.raises(SchemaYamlError, match="at least one field"):
            build_model_from_schema_yaml("{}")

    def test_unsupported_type(self) -> None:
        """未サポートの型文字列で SchemaYamlError が送出されること."""
        with pytest.raises(SchemaYamlError, match="Unsupported type 'CustomClass'"):
            build_model_from_schema_yaml("field: CustomClass")

    def test_non_string_type_value(self) -> None:
        """型の値が文字列でない場合に SchemaYamlError が送出されること."""
        with pytest.raises(SchemaYamlError, match="must be a string"):
            build_model_from_schema_yaml("field: 123")

    def test_non_string_field_name(self) -> None:
        """フィールド名が文字列でない場合に SchemaYamlError が送出されること."""
        with pytest.raises(SchemaYamlError, match="Field name must be a string"):
            build_model_from_schema_yaml("123: str")

    def test_empty_string_input(self) -> None:
        """空文字列入力で SchemaYamlError が送出されること."""
        with pytest.raises(SchemaYamlError, match="must be a YAML mapping"):
            build_model_from_schema_yaml("")

    def test_model_validate_rejects_invalid_data(self) -> None:
        """生成モデルの model_validate が不正データを拒否すること."""
        model = build_model_from_schema_yaml("name: str\nage: int")
        with pytest.raises(ValidationError):
            model.model_validate({"name": "Alice", "age": "not_an_int"})

    def test_required_field_missing(self) -> None:
        """必須フィールドが欠けている場合に ValidationError が出ること."""
        model = build_model_from_schema_yaml("name: str\nage: int")
        with pytest.raises(ValidationError):
            model.model_validate({"name": "Alice"})



"""Defines the base schema for Yagra YAML."""

from __future__ import annotations

from typing import Any

from pydantic import BaseModel, ConfigDict, Field


class NodeSpec(BaseModel):
    """Represents a node definition in the graph."""

    model_config = ConfigDict(extra="forbid")

    id: str = Field(
        min_length=1,
        description="ノードの一意識別子。edges の source/target および start_at/end_at から参照する。英数字とアンダースコアを推奨",
        examples=["translate", "summarize", "classify"],
    )
    handler: str = Field(
        min_length=1,
        description="このノードで実行するハンドラー名。組み込みハンドラー: 'llm'（テキスト出力）, 'structured_llm'（Pydantic 構造化出力）, 'streaming_llm'（ストリーミング出力）。カスタムハンドラーは Registry に登録した名前を指定する",
        examples=["llm", "structured_llm", "streaming_llm", "my_custom_handler"],
    )
    params: dict[str, Any] = Field(
        default_factory=dict,
        description="ハンドラーに渡すパラメータ辞書。llm ハンドラー: prompt_ref または prompt（必須）, model（必須）, output_key（省略時は 'output'）。structured_llm ハンドラー: 上記に加え schema_yaml（'name: str\\nage: int' 形式）。streaming_llm ハンドラー: 上記に加え stream（bool, デフォルト true）",
        examples=[
            {
                "prompt_ref": "prompts/translate.txt",
                "model": "gpt-4o-mini",
                "output_key": "translation",
            },
            {
                "prompt": {"role": "user", "content": "要約してください: {text}"},
                "model": "gpt-4o-mini",
            },
        ],
    )


class EdgeSpec(BaseModel):
    """Represents an edge definition for transitions between nodes."""

    model_config = ConfigDict(extra="forbid")

    source: str = Field(
        min_length=1,
        description="遷移元ノードの id。nodes リストに存在するノード ID を指定する",
        examples=["translate", "classify"],
    )
    target: str = Field(
        min_length=1,
        description="遷移先ノードの id。nodes リストに存在するノード ID を指定する。条件分岐なし（condition なし）の場合は必ずこのノードへ遷移する",
        examples=["summarize", "END"],
    )
    condition: str | None = Field(
        default=None,
        description="条件分岐の条件式。ノードの state['__next__'] の値と比較する文字列、または条件関数名。省略時は無条件遷移",
        examples=["approve", "reject", None],
    )


class GraphSpec(BaseModel):
    """Represents the complete Yagra YAML definition."""

    model_config = ConfigDict(extra="forbid")

    version: str = Field(
        min_length=1,
        description="YAML スキーマのバージョン。現在は '1.0' を指定する",
        examples=["1.0"],
    )
    start_at: str = Field(
        min_length=1,
        description="ワークフロー開始時に最初に実行するノードの id。nodes リストのいずれかの id を指定する",
        examples=["translate", "input_node"],
    )
    end_at: list[str] = Field(
        min_length=1,
        description="ワークフロー終了ノードの id リスト。これらのノードが完了するとワークフローが終了する。複数指定可能",
        examples=[["translate"], ["approve", "reject"]],
    )
    nodes: list[NodeSpec] = Field(
        min_length=1,
        description="ワークフローを構成するノードの定義リスト。各ノードは id, handler, params を持つ",
        examples=[
            [
                {
                    "id": "translate",
                    "handler": "llm",
                    "params": {"prompt_ref": "prompts/translate.txt", "model": "gpt-4o-mini"},
                }
            ]
        ],
    )
    edges: list[EdgeSpec] = Field(
        min_length=0,
        description="ノード間の遷移定義リスト。source から target へのエッジを列挙する。条件分岐は condition フィールドで指定する",
        examples=[
            [
                {"source": "classify", "target": "approve", "condition": "approved"},
                {"source": "classify", "target": "reject", "condition": "rejected"},
            ]
        ],
    )
    params: dict[str, Any] = Field(
        default_factory=dict,
        description="ワークフローレベルのグローバルパラメータ。現在は使用されていないが将来の拡張のために予約されている",
        examples=[{}],
    )



"""Structured output LLM handler using litellm and Pydantic.

This module provides an LLM handler that returns type-safe structured data
by parsing LLM responses with a Pydantic model.
"""

import json
import re
import time
from typing import TYPE_CHECKING, Any

from pydantic import BaseModel, ValidationError

from yagra.handlers.llm_handler import LLMHandlerCallError, LLMHandlerConfigError
from yagra.handlers.schema_builder import build_model_from_schema_yaml
from yagra.ports.outbound.node_registry import NodeHandler

if TYPE_CHECKING:
    import litellm  # type: ignore[import-not-found]
else:
    litellm = None  # type: ignore[assignment]


def create_structured_llm_handler(
    schema: type[BaseModel] | None = None,
    retry: int = 3,
    timeout: int = 30,
) -> NodeHandler:
    r"""Pydantic モデルバリデーション付きの構造化出力 LLM ハンドラーを生成する。

    litellm を介して LLM を呼び出し、レスポンスを Pydantic モデルインスタンスに
    パースするハンドラーを生成する。LLM には JSON で返却するよう指示し、
    指定されたスキーマでバリデーションを行う。

    スキーマの指定方法は 2 通り:

    1. **静的スキーマ**: ``schema`` 引数に Pydantic BaseModel サブクラスを渡す。
    2. **動的スキーマ**: ``schema`` を省略し、ワークフロー YAML の
       ``params.schema_yaml`` にフラットな ``key: type`` 形式で定義する。

    動的スキーマの YAML 例::

        name: str
        age: int
        tags: list[str]

    Args:
        schema: Pydantic BaseModel サブクラス。省略時は実行時に
            ``params["schema_yaml"]`` から動的にモデルを生成する。
        retry: API エラー時のリトライ回数（デフォルト: 3）。
        timeout: タイムアウト秒数（デフォルト: 30）。

    Returns:
        NodeHandler: (state, params) を受け取り dict を返すハンドラー関数。
            output_key の値は Pydantic モデルインスタンスになる。

    Raises:
        ImportError: litellm がインストールされていない場合。

    Examples:
        静的スキーマ指定:

        >>> from pydantic import BaseModel
        >>> class UserInfo(BaseModel):
        ...     name: str
        ...     age: int
        >>> handler = create_structured_llm_handler(schema=UserInfo, retry=3, timeout=30)

        動的スキーマ（schema_yaml を params で指定）:

        >>> handler = create_structured_llm_handler()  # schema=None
        >>> # params["schema_yaml"] = "name: str\\nage: int" で実行時に解決

        YAML 定義例:

        .. code-block:: yaml

            nodes:
              - id: "extract"
                handler: "structured_llm"
                params:
                  schema_yaml: |
                    name: str
                    age: int
                  prompt:
                    system: "Extract user info as JSON"
                    user: "{text}"
                  model:
                    provider: "openai"
                    name: "gpt-4o"
                  output_key: "user_info"
    """
    global litellm
    if litellm is None:
        try:
            import litellm
        except ImportError as e:
            msg = (
                "litellm is not installed. "
                "Install with: pip install 'yagra[llm]' or uv add --optional llm yagra"
            )
            raise ImportError(msg) from e

    def handler(state: dict[str, Any], params: dict[str, Any]) -> dict[str, Any]:
        """LLM を呼び出し、バリデーション済み Pydantic モデルインスタンスを返す。

        JSON 出力モードで LLM を呼び出し、create_structured_llm_handler に
        指定されたスキーマ、または params["schema_yaml"] から動的に生成した
        スキーマを使ってレスポンスをパースする。

        Args:
            state: ワークフロー state 辞書。
            params: ノードパラメータ（prompt, model, schema_yaml, output_key 等）。

        Returns:
            {output_key: schema_instance} 形式の辞書。
            値はバリデーション済み Pydantic モデルインスタンス。

        Raises:
            LLMHandlerConfigError: 必須パラメータが不足または不正な場合。
            LLMHandlerCallError: LLM 呼び出しまたは JSON パース/バリデーション失敗時。
            SchemaYamlError: schema_yaml のパースまたはモデル生成に失敗した場合。
        """
        # 0. スキーマ解決（静的 > 動的）
        resolved_schema: type[BaseModel]
        if schema is not None:
            resolved_schema = schema
        else:
            schema_yaml = params.get("schema_yaml")
            if not schema_yaml:
                msg = "Either 'schema' argument or 'schema_yaml' param is required"
                raise LLMHandlerConfigError(msg)
            resolved_schema = build_model_from_schema_yaml(schema_yaml)

        # 1. パラメータ抽出と検証
        prompt = params.get("prompt")
        if not isinstance(prompt, dict):
            msg = "'prompt' must be a dict with 'system' and 'user' keys"
            raise LLMHandlerConfigError(msg)

        model = params.get("model")
        if not isinstance(model, dict):
            msg = "'model' must be a dict with 'provider', 'name', and optional 'kwargs'"
            raise LLMHandlerConfigError(msg)

        provider = model.get("provider")
        model_name = model.get("name")
        if not provider or not model_name:
            msg = "'model' must have 'provider' and 'name' keys"
            raise LLMHandlerConfigError(msg)

        output_key = params.get("output_key", "output")

        # 2. プロンプトに変数を埋め込み
        system_prompt = prompt.get("system", "")
        user_prompt_template = prompt.get("user", "")

        # input_keys が明示指定されていればそちらを優先（後方互換）
        # 未指定（None）の場合はプロンプトテンプレートから {変数名} を自動抽出
        explicit_keys = params.get("input_keys")
        keys = (
            explicit_keys
            if explicit_keys is not None
            else re.findall(r"\{(\w+)\}", user_prompt_template)
        )

        # stateから入力値を取得
        input_values = {key: state.get(key, "") for key in keys}

        # {variable}形式の変数を置換
        try:
            user_prompt = user_prompt_template.format(**input_values)
        except KeyError as e:
            msg = f"Missing key in state for prompt interpolation: {e}"
            raise LLMHandlerConfigError(msg) from e

        # 3. JSON出力を促すシステムプロンプトを付加
        json_system_prompt = (
            f"{system_prompt}\n\nRespond with valid JSON only. "
            f"The JSON must conform to the following schema:\n"
            f"{json.dumps(resolved_schema.model_json_schema(), ensure_ascii=False)}"
        ).strip()

        messages = [
            {"role": "system", "content": json_system_prompt},
            {"role": "user", "content": user_prompt},
        ]

        # 4. model.kwargs の取得（response_format はデフォルトで json_object を付加）
        model_kwargs = dict(model.get("kwargs", {}))
        if "response_format" not in model_kwargs:
            model_kwargs["response_format"] = {"type": "json_object"}

        litellm_model = f"{provider}/{model_name}"

        # 5. LLM呼び出し（リトライ付き）
        last_error = None
        for attempt in range(retry):
            try:
                response = litellm.completion(
                    model=litellm_model,
                    messages=messages,
                    timeout=timeout,
                    **model_kwargs,
                )

                if not response.choices:
                    msg = "LLM returned empty response"
                    raise LLMHandlerCallError(msg)

                content = response.choices[0].message.content
                if content is None:
                    msg = "LLM returned None content"
                    raise LLMHandlerCallError(msg)

                # 6. JSONパース・Pydanticバリデーション
                try:
                    instance = resolved_schema.model_validate_json(content)
                except (ValidationError, ValueError) as e:
                    msg = f"Failed to parse LLM response as {resolved_schema.__name__}: {e}"
                    raise LLMHandlerCallError(msg) from e

                return {output_key: instance}

            except LLMHandlerCallError:
                # LLMレスポンスのエラーはリトライせず即座に送出
                raise
            except Exception as e:
                last_error = e
                if attempt < retry - 1:
                    wait_time = 2**attempt
                    time.sleep(wait_time)
                    continue
                break

        msg = f"LLM call failed after {retry} attempts: {last_error}"
        raise LLMHandlerCallError(msg) from last_error

    return handler


STRUCTURED_LLM_HANDLER_PARAMS_SCHEMA: dict = {
    "type": "object",
    "description": "create_structured_llm_handler で生成した Pydantic 構造化出力ハンドラーのパラメータ",
    "properties": {
        "prompt": {"$ref": "#/definitions/prompt_field"},
        "prompt_ref": {"$ref": "#/definitions/prompt_ref_field"},
        "model": {
            "type": "object",
            "description": "LLM モデル設定。provider と name が必須。kwargs に追加パラメータを渡せる",
            "properties": {
                "provider": {"type": "string", "examples": ["openai", "anthropic"]},
                "name": {"type": "string", "examples": ["gpt-4o-mini", "claude-opus-4-6"]},
                "kwargs": {"type": "object"},
            },
            "required": ["provider", "name"],
        },
        "input_keys": {
            "type": "array",
            "items": {"type": "string"},
            "description": "プロンプトに渡す state キーの明示指定。省略時はプロンプトテンプレートの {変数名} から自動抽出",
        },
        "output_key": {
            "type": "string",
            "description": "構造化出力を格納するステートキー名。省略時は 'output'",
            "default": "output",
        },
        "schema_yaml": {
            "type": "string",
            "description": "Pydantic モデルのフィールド定義。'フィールド名: 型' を改行区切りで記述。対応型: str, int, float, bool, list, dict",
            "examples": ["name: str\nage: int\ncity: str", "title: str\nsummary: str\ntags: list"],
        },
    },
    "required": ["model"],
    "oneOf": [
        {"required": ["prompt"]},
        {"required": ["prompt_ref"]},
    ],
}
